package gs.config.audit

import cats.effect.Ref
import cats.effect.Sync
import gs.config.ConfigName

trait ConfigManifest[F[_]]:
  /** Retrieve a snapshot of the current state of this configuration manifest.
    * This state tracks all configuration names that the caller attempted to
    * access, along with a record of whether each attempt succeeded or failed.
    *
    * @return
    *   The current state of this manifest.
    */
  def snapshot(): F[Map[ConfigName, List[ConfigQuery]]]

  /** Record a query for some [[ConfigName]] in this manifest.
    *
    * @param name
    *   The [[ConfigName]] that was queried.
    * @param query
    *   The [[ConfigQuery]] describing the result.
    * @return
    *   Side-effect indicating that the query was recorded.
    */
  def record(
    name: ConfigName,
    query: ConfigQuery
  ): F[Unit]

object ConfigManifest:

  final class Standard[F[_]: Sync] private (
    private val manifest: Ref[F, Map[ConfigName, List[ConfigQuery]]]
  ) extends ConfigManifest[F]:

    override def snapshot(): F[Map[ConfigName, List[ConfigQuery]]] =
      manifest.get

    override def record(
      name: ConfigName,
      query: ConfigQuery
    ): F[Unit] =
      manifest.update(addQueryToName(name, query, _))

    private def addQueryToName(
      name: ConfigName,
      query: ConfigQuery,
      state: Map[ConfigName, List[ConfigQuery]]
    ): Map[ConfigName, List[ConfigQuery]] =
      state.get(name) match
        case Some(queries) => state.updated(name, queries ++ List(query))
        case None          => state + (name -> List(query))

end ConfigManifest
